/*
 * $Id: bgp.h,v 1.18 1998/09/01 19:38:40 masaki Exp $
 */

#ifndef _BGP_H
#define _BGP_H

#include <mrt.h>
#include <timer.h>
#include <trace.h>
#include <io.h>
#include <schedule.h>
#include <hash.h>

#include <rib.h>
#include <bgp_proto.h>

struct _bgp_attr_t;
#define bgp_attr_t struct _bgp_attr_t

/* various settable attributes */
#define DEFAULT_PEER_NAME	"glasshouse.merit.edu"
#define DEFAULT_PEER_AS		185
#define DEFAULT_MY_AS		6503
#define DEFAULT_MY_ID		0xc66c3cb0

#define DEFAULT_DATAFILE	"/tmp/bgp.data"

#define DEFAULT_BGP_VERSION	4
#define DEFAULT_BGP4PLUS_VERSION 0

#define DEFAULT_CONNET_RETRY_INTERVAL	60
#define DEFAULT_START_INTERVAL		10
#define DEFAULT_KEEPALIVE_INTERVAL	30
#define DEFAULT_HOLDTIME_INTERVAL	180

#define BGP_MIN_HOLDTIME        20      /* What we'll accept */

/* pdu types */
#define BGP_OPEN		1
#define BGP_UPDATE		2
#define BGP_NOTIFICATION	3
#define BGP_KEEPALIVE		4

/* saving routing table to disk (mrtd) */
#define DUMP_ASCII		1
#define DUMP_BINARY		2

extern char *sbgp_states[];
extern char *sbgp_pdus[];
extern char *sbgp_events[];
extern char *s_bgp_error[];
extern char *s_bgp_error_header[];
extern char *s_bgp_error_open[];
extern char *s_bgp_error_update[];


#define BGP_TRANSPARENT_AS	0x0001
#define BGP_TRANSPARENT_NEXTHOP	0x0002
#define BGP_CONNECT_PASSIVE	0x0004	/* don't try to connect */
#define BGP_BGP4PLUS		0x0008
#define BGP_BGP4PLUS_00	BGP_BGP4PLUS	/* draft 00 */
#define BGP_BGP4PLUS_01		0x0010	/* draft 01 */
#define BGP_BGP4PLUS_AUTO	0x0020	/* auto detection */
#define BGP_EBGP_MULTIHOP	0x0040


#define MAX_BGP_PEERS 32	/* due to long's size */

typedef struct _bgp_peer_t {
    char *name;
    interface_t *bind_if;
    prefix_t *bind_addr;
    prefix_t *local_addr;	/* local address for the peer */
    /* XXX redundant? interface's address */
    prefix_t *peer_addr;	/* peer address only */
    /* XXX redundant? gateway's prefix */
    int peer_port;		/* port number in case it isn't 179 */
    u_long peer_id;		/* peer's router id */
    gateway_t *gateway;
    prefix_t *nexthop;		/* nexthop in case of multihop */
    LINKED_LIST *aliases;	/* peer's aliases */
    int sockfd;			/* sockect connected to peer */
    int sockfd_in;		/* sockect connected to peer (incoming) */
    int sockfd_listen;		/* sockect for listening incoming connection */
    int view;			/* RIB to use when sending routes */
    u_long bit;			/* id bit (used by view in mask) */
    u_long options;		/* see options above */

    int dlist_in;		/* list num for input filtering */
    int dlist_out;		/* list num for output filtering */
    int flist_in;		/* list num for input aspath filtering */
    int flist_out;		/* list num for output aspath filtering */
    int route_map_in;		/* num for input route-map */
    int route_map_out;		/* num for output route-map */

    u_short state;
    u_char code;
    u_char subcode;

    pthread_mutex_t mutex_lock;
    pthread_t self;		/* my thread ID */
    schedule_t *schedule;	/* event processing */

    /* packet parsing storage */
    u_char buffer[BGPMAXPACKETSIZE * 2];
    u_char *read_ptr;		/* current ptr of read in buffer */
    u_char *start_ptr;		/* current ptr of start of packet */
    u_char *packet;		/* pointer to the packet */

    u_char buffer_in[BGPMAXPACKETSIZE];
    u_char *read_ptr_in;	/* current ptr of read in buffer */

    /* timers */
    mtimer_t *timer_Start;
    mtimer_t *timer_ConnectRetry;
    mtimer_t *timer_KeepAlive;
    mtimer_t *timer_HoldTime;

    /* time value of intervals in seconds */
    int Start_Interval;
    int ConnectRetry_Interval;
    int KeepAlive_Interval;
    int HoldTime_Interval;

    /* tracing stuff */
    trace_t *trace;

    time_t time;
    int default_pref;
    int maximum_prefix;

    LINKED_LIST *ll_announce;
    LINKED_LIST *ll_withdraw;
    bgp_attr_t *attr;
} bgp_peer_t;

#include <view.h>

typedef struct _bgp_t {
    int my_as;			/* host byte order */
    u_long my_id;		/* network byte order */
    int cport;			/* port to send out open on */
    int lport;			/* port to listen for open on */
    int sockfd;			/* socket we are listening on */
    view_t *view;		/* view from which RIB gets routes */

    LINKED_LIST *ll_bgp_peers;
    int num_peers;		/* number of peers */
    int num_ipv4_peers;		/* number of IPv4 peers */
    int accept_all_peers;
    trace_t *trace;

    view_t *views[MAX_BGP_VIEWS];
    HASH_TABLE *attr_hash;	/* hash of bgp attribute blocks */

    schedule_t *schedule;	/* event processing */

    void (*peer_down_call_fn) ();
    void (*update_call_fn) ();
    void (*state_change_fn) ();
    rib_update_route_t rt_update_call_fn;
    void (*peer_established_call_fn) ();

    /* Since view is created dynamically,
       there may be a dump request without a real view */
    char *dump_route_form[MAX_BGP_VIEWS];
    time_t dump_route_interval[MAX_BGP_VIEWS];
    time_t dump_route_time[MAX_BGP_VIEWS];
    int	dump_route_type;	/* DUMP_ASCII or DUMP_BINARY */

    char *dump_update_form;
    time_t dump_update_interval;
    time_t dump_update_time;

    pthread_mutex_t mutex_lock;
    LINKED_LIST *ll_networks;

    u_long bits;		/* record the peers' bits */

    u_int bgp_num_active_route_head;
    u_int bgp_num_active_route_node;

#define DEFAULT_LOCAL_PREF 100
    int default_local_pref;

    time_t Default_Start_Interval;
    time_t Default_ConnectRetry_Interval;
    time_t Default_KeepAlive_Interval;
    time_t Default_HoldTime_Interval;
} bgp_t;


extern bgp_t *BGP;

enum BGP_ATTR {
    BGP_NULL = 0,
    BGP_MY_AS,
    BGP_MY_ID,
    BGP_TRACE_STRUCT,
    BGP_PEER_DOWN_FN,
    BGP_PEER_ESTABLISHED_FN,
    BGP_RECV_UPDATE_FN,
    BGP_STATE_CHANGE_FN,
    BGP_RT_UPDATE_FN,		/* route update */
    BGP_ACCEPT_ALL_PEERS,	/* start peering sesssion with anyone */
    BGP_LPORT,			/* port to listen for connections on */
    BGP_CPORT,			/* port to attempt connections to */
    BGP_USE_PORTSERVER,		/* use portserver library for listening */
    BGP_DUMP_ROUTE_FORM,	/* a file name used in dumping a route table */
    BGP_DUMP_UPDATE_FORM,	/* a file name used in dumping updates */
};


/* a convenience structure to store all prefixes with same BGP
 * attributes while we are building BGP update packets
 */
typedef struct _bgp_bucket_t {
    LINKED_LIST *ll_prefix;
    bgp_attr_t *attr;
} bgp_bucket_t;


/* in bgp_pdu.c */
int show_bgp_neighbors_errors (uii_connection_t * uii, char *peer_or_star);
int bgp_read (bgp_peer_t * peer, int sockfd, u_char * ptr, int len);
int bgp_get_pdu (bgp_peer_t * peer);
int bgp_process_update (bgp_peer_t * peer);
int bgp_process_open (bgp_peer_t * peer);
int bgp_process_notify (bgp_peer_t * peer);
int bgp_send_open (bgp_peer_t * peer);
int bgp_send_keepalive (bgp_peer_t * peer);
int bgp_send_notification (bgp_peer_t * peer, int code, int subcode);
int bgp_send_notification2 (int sockfd, prefix_t * prefix, int port,
			    int code, int subcode);
int bgp_send_notify_byte (bgp_peer_t * peer, int code, int subcode, int opt);
int bgp_send_notify_word (bgp_peer_t * peer, int code, int subcode, int opt);
int bgp_send_update (bgp_peer_t * peer, int len, u_char * data);

/* in bgp_util.c */
char *bgp_notify_to_string (int code, int subcode);
int bgp_start_transport_connection (bgp_peer_t * peer);
void bgp_release_resources (bgp_peer_t * peer);
bgp_peer_t *Find_BGP_Peer (prefix_t * prefix, prefix_t *usrc, 
			   interface_t *interface);
bgp_peer_t *Find_BGP_Peer_ByID (char *name);
bgp_peer_t *Add_BGP_Peer (char *name, prefix_t * prefix, prefix_t *usrc,
			  interface_t *interface, int as, trace_t * trace);
void Destroy_BGP_Peer (bgp_peer_t * peer);
void bgp_flush_socket (bgp_peer_t * peer);
void bgp_broken_pipe ();
void bgp_die ();
int bgp_kill_peer (gateway_t * gateway);
int bgp_kill_peer2 (bgp_peer_t * peer);
void bgp_kill_all ();
int default_bgp_peer_down_fd (bgp_peer_t * peer);
void bgp_start_peer_thread (bgp_peer_t * peer);
void bgp_start_main_thread ();
bgp_peer_t *find_bgp_peer (gateway_t * gateway);
int show_bgp (uii_connection_t * uii);
int show_bgp_routing_table (uii_connection_t * uii);
int show_bgp_rt_regexp (uii_connection_t * uii, char *expr);
void bgp_peer_dead (bgp_peer_t * peer);
void bgp_dump_view (uii_connection_t * uii, view_t *view, 
		    as_regexp_code_t * code, bgp_peer_t *peer);
int set_BGP (int first,...);
int start_bgp (void);
void start_bgp_peer (bgp_peer_t * peer);
int trace_bgp (uii_connection_t * uii);
int no_trace_bgp (uii_connection_t * uii);
int show_bgp_neighbors_routes (uii_connection_t * uii, char *peer_or_star);
int bgp_check_attr (bgp_peer_t * peer, bgp_attr_t * attr, int as);
int check_bgp_networks (prefix_t *prefix);

/* in bgp_sm.c */
void bgp_change_state (bgp_peer_t * peer, int state, int event);
void bgp_sm_state_idle (bgp_peer_t * peer, int event);
void bgp_sm_state_active (bgp_peer_t * peer, int event);
void bgp_sm_state_connect (bgp_peer_t * peer, int event);
void bgp_sm_state_opensent (bgp_peer_t * peer, int event);
void bgp_sm_state_openconfirm (bgp_peer_t * peer, int event);
void bgp_sm_state_established (bgp_peer_t * peer, int event);
void bgp_sm_process_event (bgp_peer_t * peer, int event);

/* in bgp_timer.c */
void bgp_timer_ConnectRetry_fire (mtimer_t * timer, bgp_peer_t * peer);
void bgp_timer_KeepAlive_fire (mtimer_t * timer, bgp_peer_t * peer);
void bgp_timer_HoldTime_fire (mtimer_t * timer, bgp_peer_t * peer);
void bgp_timer_StartTime_fire (mtimer_t * timer, bgp_peer_t * peer);

/* in bgp_thread.c */
void bgp_schedule_timer (mtimer_t * timer, bgp_peer_t * peer);
void bgp_schedule_socket (bgp_peer_t * peer);
void bgp_get_config_neighbor (bgp_peer_t * peer);
void bgp_schedule_get_config_neighbor (bgp_peer_t * peer);

/* public functions */
int init_BGP (trace_t * ltrace);

/* bgp_dump.c */
void bgp_write_mrt_msg (bgp_peer_t * peer, enum MRT_MSG_BGP_TYPES bgptype,
			u_char * buf, int length);
int bgp_write_status_change (bgp_peer_t * peer, u_short state);

#define Find_BGP_Peer_ByPrefix(prefix) Find_BGP_Peer(prefix, NULL, NULL)

#undef bgp_attr_t
#endif /* _BGP_H */
